# Generated by Django 3.2.9 on 2022-03-31 20:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='donationproject',
            name='name',
            field=models.CharField(max_length=50),
        ),
    ]



from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import Http404
from django.urls import reverse
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.models import DonationBlueprint, ZippedBlueprint, DonationInstruction


class BlueprintMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        return context

    def get_success_url(self):
        return reverse('blueprint-list', kwargs={'project_pk': self.kwargs['project_pk']})


class ProjectBlueprintList(LoginRequiredMixin, BlueprintMixin, ListView):
    """ View to list all donation blueprints associated with a project. """
    model = DonationBlueprint
    context_object_name = 'donation_blueprints'
    template_name = 'ddm/project_admin/blueprint/list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({
            'zipped_blueprints': ZippedBlueprint.objects.filter(project_id=self.kwargs['project_pk']),
        })
        return context

    def get_queryset(self):
        queryset = super().get_queryset().filter(project_id=self.kwargs['project_pk'])
        return queryset


class BlueprintCreate(LoginRequiredMixin, BlueprintMixin, CreateView):
    """ View to create a new donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/create.html'
    fields = ['name', 'exp_file_format', 'expected_fields', 'extracted_fields']

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)


class BlueprintEdit(LoginRequiredMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/edit.html'
    fields = [
        'name', 'exp_file_format', 'expected_fields', 'extracted_fields',
        'zip_blueprint', 'regex_path'
    ]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['form'].fields['zip_blueprint'].queryset = ZippedBlueprint.objects.filter(project_id=self.kwargs['project_pk'])
        return context


class BlueprintDelete(LoginRequiredMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing donation blueprint. """
    model = DonationBlueprint
    template_name = 'ddm/project_admin/blueprint/delete.html'


class ZippedBlueprintCreate(LoginRequiredMixin, BlueprintMixin, CreateView):
    """ View to create a new zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/create.html'
    fields = ['name']

    def form_valid(self, form):
        form.instance.project_id = self.kwargs['project_pk']
        return super().form_valid(form)


class ZippedBlueprintEdit(LoginRequiredMixin, BlueprintMixin, UpdateView):
    """ View to edit the details of an existing zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/edit.html'
    fields = ['name']


class ZippedBlueprintDelete(LoginRequiredMixin, BlueprintMixin, DeleteView):
    """ View to delete an existing zipped blueprint. """
    model = ZippedBlueprint
    template_name = 'ddm/project_admin/blueprint/delete.html'


class InstructionMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        if self.kwargs['blueprint_type'] == 'blueprint':
            blueprint = DonationBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
            blueprint_type = 'blueprint'
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            blueprint = ZippedBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
            blueprint_type = 'zipped Blueprint'
        else:
            raise Http404
        context.update({
            'blueprint': blueprint,
            'blueprint_type': blueprint_type,
        })
        return context

    def get_success_url(self):
        kwargs = {'project_pk': self.kwargs['project_pk'],
                  'blueprint_type': self.kwargs['blueprint_type'],
                  'blueprint_pk': self.kwargs['blueprint_pk']}
        return reverse('instruction-overview', kwargs=kwargs)


class InstructionOverview(LoginRequiredMixin, InstructionMixin, ListView):
    """ View to create a new instruction page. """
    model = DonationInstruction
    context_object_name = 'instructions'
    template_name = 'ddm/project_admin/instructions/list.html'
    fields = ['text', 'index']

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.kwargs['blueprint_type'] == 'blueprint':
            queryset = queryset.filter(blueprint_id=self.kwargs['blueprint_pk'])
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            queryset = queryset.filter(zip_blueprint_id=self.kwargs['blueprint_pk'])
        else:
            raise Http404
        return queryset


class InstructionCreate(LoginRequiredMixin, InstructionMixin, CreateView):
    """ View to create an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/create.html'
    fields = ['text', 'index']

    def get_blueprint(self):
        if self.kwargs['blueprint_type'] == 'blueprint':
            blueprint = DonationBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
        elif self.kwargs['blueprint_type'] == 'zip-blueprint':
            blueprint = ZippedBlueprint.objects.get(id=self.kwargs['blueprint_pk'])
        return blueprint

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        blueprint = self.get_blueprint()
        if isinstance(blueprint, DonationBlueprint):
            kwargs['instance'] = DonationInstruction(blueprint=blueprint)
        elif isinstance(blueprint, ZippedBlueprint):
            kwargs['instance'] = DonationInstruction(zip_blueprint=blueprint)
        return kwargs

    def get_initial(self):
        initial = super().get_initial()
        indices = self.get_blueprint().donationinstruction_set.values_list('index', flat=True)
        if indices:
            initial['index'] = max(indices) + 1
        else:
            initial['index'] = 1
        return initial


class InstructionEdit(LoginRequiredMixin, InstructionMixin, UpdateView):
    """ View to edit an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/edit.html'
    fields = ['text', 'index']


class InstructionDelete(LoginRequiredMixin, InstructionMixin, DeleteView):
    """ View to delete an instruction page. """
    model = DonationInstruction
    template_name = 'ddm/project_admin/instructions/delete.html'

    def get_success_url(self):
        kwargs = {'project_pk': self.kwargs['project_pk'],
                  'blueprint_type': self.kwargs['blueprint_type'],
                  'blueprint_pk': self.kwargs['blueprint_pk']}
        return reverse('instruction-overview', kwargs=kwargs)



# Generated by Django 3.2.9 on 2022-04-29 09:54

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0002_alter_donationproject_name'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='datadonation',
            name='time',
        ),
        migrations.AddField(
            model_name='datadonation',
            name='time_submitted',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='date_created',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='donationproject',
            name='public_key',
            field=models.TextField(default=1234),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='donationproject',
            name='super_secret',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='datadonation',
            name='data',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='questionnaireresponse',
            name='data',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='questionnaireresponse',
            name='time_submitted',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView

from ddm.models import DonationProject, DonationBlueprint, ZippedBlueprint


class ProjectList(LoginRequiredMixin, ListView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/list.html'


class ProjectCreate(LoginRequiredMixin, CreateView):
    """ View to create a new donation project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/create.html'
    fields = ['name', 'slug']


class ProjectDetail(LoginRequiredMixin, DetailView):
    """ View to display landing page for project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/detail.html'


class ProjectEdit(LoginRequiredMixin, UpdateView):
    """ View to edit the details of an existing donation project. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit.html'
    fields = ['name', 'slug']


class ProjectDelete(LoginRequiredMixin, DeleteView):
    """ View to display a list of existing donation projects. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/delete.html'
    success_url = reverse_lazy('project-list')


class WelcomePageEdit(LoginRequiredMixin, UpdateView):
    """ View to edit the welcome page. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit.html'
    fields = ['intro_text']


class EndPageEdit(LoginRequiredMixin, UpdateView):
    """ View to edit the welcome page. """
    model = DonationProject
    template_name = 'ddm/project_admin/project/edit.html'
    fields = ['outro_text']



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse
from ddm.models import (
    DonationProject, DonationBlueprint, OpenQuestion, ZippedBlueprint,
    SingleChoiceQuestion, MatrixQuestion, DonationInstruction
)


class BaseTestCaseAdminViews:
    @classmethod
    def setUpTestData(cls):
        cls.project = DonationProject.objects.create(
            name='Test Project',
            slug='test-project'
        )
        cls.dbp = DonationBlueprint.objects.create(
            project=cls.project,
            name='donation blueprint',
            expected_fields='a,b',
            extracted_fields='a'
        )
        cls.zbp = ZippedBlueprint.objects.create(
            name='zipped blueprint',
            project=cls.project
        )
        cls.open_question = OpenQuestion.objects.create(
            project=cls.project,
            blueprint=cls.dbp,
            name='open question',
            variable_name='open_question'
        )
        cls.sc_question = SingleChoiceQuestion.objects.create(
            project=cls.project,
            blueprint=cls.dbp,
            name='sc question',
            variable_name='sc_question'
        )
        cls.matrix_question = MatrixQuestion.objects.create(
            project=cls.project,
            blueprint=cls.dbp,
            name='matrix question',
            variable_name='matrix_question'
        )
        cls.instruction = DonationInstruction.objects.create(
            text='some text',
            index=1,
            blueprint=cls.dbp
        )
        cls.user_name = 'test-user'
        cls.user_pw = 'test-password'
        User.objects.create_superuser(cls.user_name, 'test@test.com', cls.user_pw)

    def test_logged_in_returns_200(self):
        self.client.login(username=self.user_name, password=self.user_pw)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

    def test_logged_out_returns_404(self):
        response = self.client.get(self.url, follow=True)
        self.assertEqual(response.status_code, 404)


class ProjectListViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-list', args=[])


class ProjectCreateViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-create', args=[])


class ProjectEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-edit', args=[self.project.pk])


class ProjectDeleteViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('project-delete', args=[self.project.pk])


class ProjectWelcomeEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('welcome-page-edit', args=[self.project.pk])


class ProjectEndEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('end-page-edit', args=[self.project.pk])


class BlueprintListViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-list', args=[self.project.pk])


class BlueprintCreateViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-create', args=[self.project.pk])


class BlueprintEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-edit', args=[self.project.pk, self.dbp.pk])


class BlueprintDeleteViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('blueprint-delete', args=[self.project.pk, self.dbp.pk])


class ZippedBlueprintCreateViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-create', args=[self.project.pk])


class ZippedBlueprintEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-edit', args=[self.project.pk, self.zbp.pk])


class ZippedBlueprintDeleteViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('zipped-blueprint-delete', args=[self.project.pk, self.zbp.pk])


class InstructionListViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'instruction-overview',
            args=[self.project.pk, 'blueprint', self.dbp.pk]
        )


class InstructionCreateViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'instruction-create',
            args=[self.project.pk, 'blueprint', self.dbp.pk]
        )


class InstructionEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'instruction-edit',
            args=[self.project.pk, 'blueprint', self.dbp.pk, self.instruction.pk]
        )


class InstructionDeleteViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'instruction-delete',
            args=[self.project.pk, 'blueprint', self.dbp.pk, self.instruction.pk]
        )


class QuestionnaireOverviewViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse('questionnaire-overview',  args=[self.project.pk])


class QuestionCreateViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.question_type_slugs = [
            'single_choice',
            'multi_choice',
            'open',
            'matrix',
            'semantic_diff',
            'transition'
        ]

    def test_logged_in_returns_200(self):
        self.client.login(username=self.user_name, password=self.user_pw)
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project.pk, slug])
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)

    def test_logged_out_returns_404(self):
        for slug in self.question_type_slugs:
            url = reverse('question-create', args=[self.project.pk, slug])
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 404)


class QuestionEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'question-edit',
            args=[self.project.pk, 'open', self.open_question.pk]
        )


class QuestionDeleteViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'question-delete',
            args=[self.project.pk, 'open', self.open_question.pk]
        )


class QuestionItemsEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'question-items',
            args=[self.project.pk, 'single_choice', self.sc_question.pk]
        )


class QuestionScaleEditViewTests(BaseTestCaseAdminViews, TestCase):
    def setUp(self):
        super().setUp()
        self.url = reverse(
            'question-scale',
            args=[self.project.pk, 'matrix', self.matrix_question.pk]
        )



from django.urls import reverse
from ckeditor.fields import RichTextField
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.db.models.signals import pre_save
from django.dispatch import receiver

from ddm.models import Encryption


class DonationProject(models.Model):
    name = models.CharField(max_length=50)
    slug = models.SlugField(
        verbose_name='External Project Slug',
        unique=True
    )
    intro_text = RichTextField(null=True, blank=True, verbose_name="Welcome Page Text")
    outro_text = RichTextField(null=True, blank=True, verbose_name="End Page Text")

    date_created = models.DateTimeField(default=timezone.now)

    secret_key = settings.SECRET_KEY
    public_key = models.BinaryField()
    super_secret = models.BooleanField(default=False)

    # owner = None  # TODO: Add FK to Owner.

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('project-detail', args=[str(self.id)])

    @property
    def secret(self):
        return self.secret_key

    @secret.setter
    def secret(self, value):
        self.secret_key = value


@receiver(pre_save, sender=DonationProject)
def my_callback(sender, instance, *args, **kwargs):
    if not instance.public_key:
        instance.public_key = Encryption(
            instance.secret_key, str(instance.date_created)).public_key
        instance.super_secret = True


class Participant(models.Model):
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )

    # Participation statistics.
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(null=True)
    completed = models.BooleanField(default=False)


class QuestionnaireResponse(models.Model):
    # Will only ever be deleted, when the project is deleted.
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )
    participant = models.ForeignKey(
        Participant,
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    data = models.BinaryField()

    def save(self, *args, **kwargs):
        self.data = Encryption(
            public_key=self.project.public_key
        ).encrypt(self.data)
        super().save(*args, **kwargs)

    def get_decrypted_data(self, secret=None):
        if not secret:
            decrypted_data = Encryption(
                secret=settings.SECRET_KEY,
                salt=str(self.project.date_created)
            ).decrypt(self.data)
        else:
            decrypted_data = Encryption(
                secret=secret,
                salt=str(self.project.date_created)
            ).decrypt(self.data)
        return decrypted_data



from .encryption import Encryption
from .projects import DonationProject, Participant, QuestionnaireResponse
from .data_donations import (
    DonationBlueprint, ZippedBlueprint, DataDonation, DonationInstruction
)
from .questions import *



from django.test import TestCase, Client
from django.urls import reverse
from django.utils import timezone
from ddm.models import (
    DonationProject, DonationBlueprint, OpenQuestion, DataDonation, Participant
)


class ParticipationFlowBaseTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Project with associated data donation and questions:
        cls.project = DonationProject.objects.create(
            name='Test Project',
            slug='test-project'
        )
        cls.dbp = DonationBlueprint.objects.create(
            project=cls.project,
            name='donation blueprint',
            expected_fields='a,b',
            extracted_fields='a'
        )
        OpenQuestion.objects.create(
            project=cls.project,
            blueprint=cls.dbp,
            name='a question',
            variable_name='a_question'
        )
        # Project without associated data donation and questions:
        cls.project_no = DonationProject.objects.create(
            name='Test Project without Questionnaire',
            slug='test-project-without'
        )

    def setUp(self):
        # URLs for project with questionnaire.
        self.entry_url = reverse('project-entry', args=[self.project.slug])
        self.dd_url = reverse('data-donation', args=[self.project.slug])
        self.quest_url = reverse('questionnaire', args=[self.project.slug])
        self.exit_url = reverse('project-exit', args=[self.project.slug])

        # URLs for project without questionnaire.
        self.entry_url_no = reverse('project-entry', args=[self.project_no.slug])
        self.dd_url_no = reverse('data-donation', args=[self.project_no.slug])
        self.quest_url_no = reverse('questionnaire', args=[self.project_no.slug])
        self.exit_url_no = reverse('project-exit', args=[self.project_no.slug])

        # URLs for non-existing project.
        self.entry_url_invalid = reverse('project-entry', args=['nope'])
        self.dd_url_invalid = reverse('data-donation', args=['nope'])
        self.quest_url_invalid = reverse('questionnaire', args=['nope'])
        self.exit_url_invalid = reverse('project-exit', args=['nope'])

    def initialize_project_and_session(self):
        self.client = Client()
        self.client.get(self.entry_url)
        self.client.get(self.entry_url_no)

    def create_data_donation(self):
        # Create a data donation for project 1.
        participant_id = self.client.session['projects'][f'{self.project.pk}']['participant_id']
        participant = Participant.objects.get(pk=int(participant_id))
        DataDonation.objects.create(
            project=self.project,
            blueprint=self.dbp,
            participant=participant,
            time_submitted=timezone.now(),
            consent=True,
            status='{}',
            data='{}'
        )


class TestEntryView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()

    def test_project_entry_view_registers_project(self):
        self.client.get(self.entry_url)
        session = self.client.session
        expected_keys = ['steps', 'data', 'completed', 'participant_id']
        assert set(expected_keys).issubset(set(session['projects'][f'{self.project.pk}'].keys()))

    def test_project_entry_view_registers_participant(self):
        nr_participants_before = len(Participant.objects.all())
        self.client.get(self.entry_url)
        self.assertGreater(len(Participant.objects.all()), nr_participants_before)
        self.assertIsNotNone(self.client.session['projects'][f'{self.project.pk}']['participant_id'])

    def test_project_entry_view_GET_valid_url(self):
        response = self.client.get(self.entry_url)
        project_session = self.client.session['projects'][f'{self.project.pk}']
        self.assertEqual(project_session['steps']['project-entry']['state'], 'started')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/entry_page.html')

    def test_project_entry_view_GET_invalid_url(self):
        response = self.client.get(self.entry_url_invalid)
        self.assertEqual(response.status_code, 404)

    def test_project_entry_POST(self):
        response = self.client.post(self.entry_url)
        project_session = self.client.session['projects'][f'{self.project.pk}']
        self.assertEqual(project_session['steps']['project-entry']['state'], 'completed')
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.dd_url)


class TestDonationView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()
        session = self.client.session
        session['projects'][f'{self.project.pk}']['steps']['project-entry']['state'] = 'completed'
        session.save()

    def test_data_donation_GET_valid_url(self):
        response = self.client.get(self.dd_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/data_donation.html')

    def test_data_donation_GET_invalid_url(self):
        response = self.client.get(self.dd_url_invalid)
        self.assertEqual(response.status_code, 404)

    def test_data_donation_POST_redirect(self):
        response = self.client.post(self.dd_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.quest_url)


class TestQuestionnaireView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        self.create_data_donation()
        session = self.client.session
        session['projects'][f'{self.project.pk}']['steps']['project-entry']['state'] = 'completed'
        session['projects'][f'{self.project.pk}']['steps']['data-donation']['state'] = 'completed'
        session['projects'][f'{self.project_no.pk}']['steps']['project-entry']['state'] = 'completed'
        session['projects'][f'{self.project_no.pk}']['steps']['data-donation']['state'] = 'completed'
        session.save()

    def test_questionnaire_GET_valid_url(self):
        response = self.client.get(self.quest_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/questionnaire.html')

    def test_questionnaire_GET_invalid_url(self):
        response = self.client.get(self.quest_url_invalid)
        self.assertEqual(response.status_code, 404)

    def test_questionnaire_GET_no_questionnaire(self):
        good_response = self.client.get(self.quest_url_no)
        self.assertRedirects(good_response, self.exit_url_no)

    def test_questionnaire_POST_redirect(self):
        response = self.client.post(self.quest_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, self.exit_url)


class TestExitView(ParticipationFlowBaseTestCase):

    def setUp(self):
        super().setUp()
        self.initialize_project_and_session()
        session = self.client.session
        session['projects'][f'{self.project.pk}']['steps']['project-entry']['state'] = 'completed'
        session['projects'][f'{self.project.pk}']['steps']['data-donation']['state'] = 'completed'
        session['projects'][f'{self.project.pk}']['steps']['questionnaire']['state'] = 'completed'
        session.save()

    def test_project_exit_GET_valid_url(self):
        response = self.client.get(self.exit_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/end.html')

    def test_project_exit_GET_invalid_url(self):
        response = self.client.get(self.exit_url_invalid)
        self.assertEqual(response.status_code, 404)

    def test_project_exit_POST(self):
        response = self.client.post(self.exit_url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'ddm/public/end.html')


# TODO: Add Class TestViewsRerouting(TestCase)



from django.urls import path, include

from ddm.views import project_admin
from ddm.views import participation_flow

participation_flow_patterns = [
    path(r'intro/', participation_flow.EntryView.as_view(), name='project-entry'),
    path(r'data-donation/', participation_flow.DataDonationView.as_view(), name='data-donation'),
    path(r'questionnaire/', participation_flow.QuestionnaireView.as_view(), name='questionnaire'),
    path(r'end/', participation_flow.ExitView.as_view(), name='project-exit')
]

question_patterns = [
    path(r'', project_admin.QuestionnaireOverview.as_view(), name='questionnaire-overview'),
    path(r'<slug:question_type>/create/', project_admin.QuestionCreate.as_view(), name='question-create'),
    path(r'<slug:question_type>/<int:pk>/edit/', project_admin.QuestionEdit.as_view(), name='question-edit'),
    path(r'<slug:question_type>/<int:pk>/delete/', project_admin.QuestionDelete.as_view(), name='question-delete'),
    path(r'<slug:question_type>/<int:pk>/items/', project_admin.ItemEdit.as_view(), name='question-items'),
    path(r'<slug:question_type>/<int:pk>/scale/', project_admin.ScaleEdit.as_view(), name='question-scale'),
]

instruction_patterns = [
    path(r'', project_admin.InstructionOverview.as_view(), name='instruction-overview'),
    path(r'create/', project_admin.InstructionCreate.as_view(), name='instruction-create'),
    path(r'<int:pk>/edit/', project_admin.InstructionEdit.as_view(), name='instruction-edit'),
    path(r'<int:pk>/delete/', project_admin.InstructionDelete.as_view(), name='instruction-delete'),
]

blueprint_patterns = [
    path(r'', project_admin.data_donations.ProjectBlueprintList.as_view(), name='blueprint-list'),
    path(r'blueprint/create/', project_admin.BlueprintCreate.as_view(), name='blueprint-create'),
    path(r'blueprint/<int:pk>/edit/', project_admin.BlueprintEdit.as_view(), name='blueprint-edit'),
    path(r'blueprint/<int:pk>/delete/', project_admin.BlueprintDelete.as_view(), name='blueprint-delete'),
    path(r'zip-blueprint/create/', project_admin.ZippedBlueprintCreate.as_view(), name='zipped-blueprint-create'),
    path(r'zip-blueprint/<int:pk>/edit/', project_admin.ZippedBlueprintEdit.as_view(), name='zipped-blueprint-edit'),
    path(r'zip-blueprint/<int:pk>/delete/', project_admin.ZippedBlueprintDelete.as_view(), name='zipped-blueprint-delete'),
    path(r'<slug:blueprint_type>/<int:blueprint_pk>/instructions/', include(instruction_patterns)),
]

project_admin_patterns = [
    path(r'', project_admin.ProjectList.as_view(), name='project-list'),
    path(r'create/', project_admin.ProjectCreate.as_view(), name='project-create'),
    path(r'<int:pk>/', project_admin.ProjectDetail.as_view(), name='project-detail'),
    path(r'<int:pk>/edit/', project_admin.ProjectEdit.as_view(), name='project-edit'),
    path(r'<int:pk>/delete/', project_admin.ProjectDelete.as_view(), name='project-delete'),
    path(r'<int:pk>/welcome-page/', project_admin.WelcomePageEdit.as_view(), name='welcome-page-edit'),
    path(r'<int:pk>/end-page/', project_admin.EndPageEdit.as_view(), name='end-page-edit'),
    path(r'<int:project_pk>/questionnaire/', include(question_patterns)),
    path(r'<int:project_pk>/donation-blueprints/', include(blueprint_patterns)),
]

urlpatterns = [
    path(r'<slug:slug>/', include(participation_flow_patterns)),
    path(r'projects/', include(project_admin_patterns))
]



from django.contrib.auth.mixins import LoginRequiredMixin
from django.forms import inlineformset_factory
from django.urls import reverse
from django.views.generic.edit import CreateView, UpdateView, DeleteView

from ddm.models import (
    QuestionBase, QuestionType, SingleChoiceQuestion, MultiChoiceQuestion,
    OpenQuestion, MatrixQuestion, SemanticDifferential, Transition, QuestionItem,
    ScalePoint, DonationBlueprint
)
from . import ProjectBlueprintList


class ProjectMixin:
    """ Mixin for all blueprint related views. """
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'project_pk': self.kwargs['project_pk']})
        return context


class QuestionnaireOverview(ProjectBlueprintList):
    """ View to list all donation blueprints associated with a project. """
    template_name = 'ddm/project_admin/questionnaire/list.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_types = QuestionType.choices
        context.update({'question_types': question_types})
        return context


class QuestionFormMixin(ProjectMixin):
    model = None
    fields = '__all__'

    QUESTION_CLASSES = {
        'single_choice': SingleChoiceQuestion,
        'multi_choice': MultiChoiceQuestion,
        'open': OpenQuestion,
        'matrix': MatrixQuestion,
        'semantic_diff': SemanticDifferential,
        'transition': Transition
    }

    SHARED_FIELDS = ['name', 'blueprint', 'index', 'variable_name', 'text', 'required']
    QUESTION_FIELDS = {
        'single_choice': SHARED_FIELDS + ['randomize_items'],
        'multi_choice': SHARED_FIELDS + ['randomize_items'],
        'matrix': SHARED_FIELDS + ['randomize_items'],
        'semantic_diff': SHARED_FIELDS + ['randomize_items'],
        'open': SHARED_FIELDS,
        'transition': SHARED_FIELDS
    }

    def get_queryset(self):
        self.model = self.QUESTION_CLASSES[self.kwargs['question_type']]
        self.fields = self.QUESTION_FIELDS[self.kwargs['question_type']]
        return super().get_queryset()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        question_type_label = QuestionType(self.kwargs['question_type']).label
        context.update({'question_type': question_type_label})
        context['form'].fields['blueprint'].queryset = DonationBlueprint.objects.filter(project_id=self.kwargs['project_pk'])
        return context


class QuestionCreate(LoginRequiredMixin, QuestionFormMixin, CreateView):
    """ View to create question. """
    template_name = 'ddm/project_admin/questionnaire/create.html'

    def get_initial(self):
        initial = super().get_initial()
        initial['question_type'] = self.kwargs['question_type']
        return initial

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.QUESTION_CLASSES[self.kwargs['question_type']](
            project_id=self.kwargs['project_pk'],
        )
        return kwargs

    def get_success_url(self):
        return reverse('questionnaire-overview', kwargs={'project_pk': self.kwargs['project_pk']})


class QuestionEdit(LoginRequiredMixin, QuestionFormMixin, UpdateView):
    """ View to edit question. """
    model = QuestionBase
    template_name = 'ddm/project_admin/questionnaire/edit.html'
    fields = '__all__'

    def get_success_url(self):
        success_kwargs = {
            'project_pk': self.kwargs['project_pk'],
            'question_type': self.kwargs['question_type'],
            'pk': self.kwargs['pk']
        }
        return reverse('question-edit', kwargs=success_kwargs)


class QuestionDelete(LoginRequiredMixin, ProjectMixin, DeleteView):
    """ View to delete question. """
    model = QuestionBase
    template_name = 'ddm/project_admin/questionnaire/delete.html'

    def get_success_url(self):
        return reverse('questionnaire-overview', kwargs={'project_pk': self.kwargs['project_pk']})


class InlineFormsetMixin(ProjectMixin):
    model = QuestionBase
    fields = '__all__'
    formset_model = None
    context_title = None
    fields_to_exclude = ()

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        formset = self.get_formset()
        if formset.is_valid():
            return self.form_valid(formset)
        else:
            return super().form_invalid(formset)

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        context.update({
            'formset': self.get_formset(),
            'context_title': self.context_title,
            'question': self.object
        })
        return context

    def get_formset(self):
        formset = inlineformset_factory(
            QuestionBase, self.formset_model, exclude=self.get_excluded_fields(),
            extra=1
        )
        return formset(self.request.POST or None, instance=self.object)

    def get_excluded_fields(self):
        if not isinstance(self.object, SemanticDifferential):
            excluded_fields = ('label_alt',)
        else:
            excluded_fields = ()
        return excluded_fields

    def get_success_url(self):
        question = self.get_object()
        success_kwargs = {
            'project_pk': self.kwargs['project_pk'],
            'question_type': question.question_type,
            'pk': question.pk
        }
        return reverse('question-edit', kwargs=success_kwargs)


class ItemEdit(LoginRequiredMixin, InlineFormsetMixin, UpdateView):
    """ View to edit the items associated with a question. """
    model = QuestionBase
    formset_model = QuestionItem
    template_name = 'ddm/project_admin/questionnaire/edit_set.html'
    context_title = 'Items'


class ScaleEdit(LoginRequiredMixin, InlineFormsetMixin, UpdateView):
    """ View to edit the scale associated with a question. """
    model = QuestionBase
    formset_model = ScalePoint
    template_name = 'ddm/project_admin/questionnaire/edit_set.html'
    context_title = 'Scale Points'



from ckeditor.fields import RichTextField
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from django.utils import timezone

from ddm.models import DonationProject, Participant, Encryption

import json

import logging
logger = logging.getLogger(__name__)


class ZippedBlueprint(models.Model):
    name = models.CharField(max_length=250)
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )

    def __str__(self):
        return self.name

    def get_slug(self):
        return 'zip-blueprint'

    def get_absolute_url(self):
        return reverse('zipped-blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_blueprints(self):
        blueprints = DonationBlueprint.objects.filter(zip_blueprint=self)
        return blueprints

    def get_configs(self):
        bp_configs = []
        blueprints = self.get_blueprints()
        for bp in blueprints:
            bp_configs.append(bp.get_config())
        return bp_configs

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]


# TODO: For admin section: Add validation on save to ensure that regex_path != None when zip_blueprint != None
class DonationBlueprint(models.Model):
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )

    name = models.CharField(max_length=250)

    class FileFormats(models.TextChoices):
        JSON_FORMAT = 'json'
        # CSV_FORMAT = 'csv',
        # HTML_FORMAT = 'html',
        # XLSX_FORMAT = 'xlsx',

    exp_file_format = models.CharField(
        max_length=10,
        choices=FileFormats.choices,
        default=FileFormats.JSON_FORMAT,
        verbose_name='Expected File Format'
    )

    expected_fields = models.JSONField()
    extracted_fields = models.JSONField()

    # Configuration if related to ZippedBlueprint:
    zip_blueprint = models.ForeignKey(
        ZippedBlueprint,
        null=True,
        blank=True,
        on_delete=models.SET_NULL
    )
    regex_path = models.TextField(
        null=True,
        blank=True
    )

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('blueprint-edit', args=[str(self.project_id), str(self.id)])

    def get_slug(self):
        return 'blueprint'

    def get_config(self):
        config = {
            'id': self.pk,
            'name': self.name,
            'format': self.exp_file_format,
            'f_expected': self.expected_fields,
            'f_extract': self.extracted_fields,
            'regex_path': self.regex_path,
        }
        return config

    def get_instructions(self):
        return [{'index': i.index, 'text': i.text} for i in self.donationinstruction_set.all()]

    def get_associated_questions(self):
        return self.questionbase_set.all()

    def process_donation(self, data, participant):
        if self.validate_donation(data):
            self.create_donation(data, participant)
        else:
            logger.error(f'Donation not processed by blueprint {self.pk}')
        return

    @staticmethod
    def validate_donation(data):
        # Check if all expected fields are in response.
        response_fields = ['consent', 'extracted_data', 'status']
        if not all(k in data for k in response_fields):
            logger.error(
                'Donation data does not contain the expected information. '
                f'Expected key: {response_fields}; '
                f'Present fields: {data.keys()}.'
            )
            return False

        # TODO: Add other validation steps? - e.g., validation of extracted fields
        return True

    def create_donation(self, data, participant):
        DataDonation.objects.create(
            project=self.project,
            blueprint=self,
            participant=participant,
            consent=data['consent'],
            status=data['status'],
            data=data['extracted_data']
        )
        return


class DataDonation(models.Model):
    project = models.ForeignKey(
        DonationProject,
        on_delete=models.CASCADE
    )
    blueprint = models.ForeignKey(
        DonationBlueprint,
        null=True,
        on_delete=models.SET_NULL
    )
    participant = models.ForeignKey(
        Participant,
        on_delete=models.CASCADE
    )
    time_submitted = models.DateTimeField(default=timezone.now)
    consent = models.BooleanField(default=False)
    status = models.JSONField()
    data = models.BinaryField()

    def save(self, *args, **kwargs):
        # TODO: Enable encryption again:
        # self.data = Encryption(
        #     public_key=self.project.public_key
        # ).encrypt(self.data)
        self.data = json.dumps(self.data).encode('utf8')
        super().save(*args, **kwargs)

    def get_decrypted_data(self, secret=None):
        # TODO: Enable encryption again:
        # if not secret:
        #     decrypted_data = Encryption(
        #         secret=settings.SECRET_KEY,
        #         salt=str(self.project.date_created)
        #     ).decrypt(self.data)
        # else:
        #     decrypted_data = Encryption(
        #         secret=secret,
        #         salt=str(self.project.date_created)
        #     ).decrypt(self.data)
        decrypted_data = json.loads(self.data.decode('utf8'))
        return decrypted_data


class DonationInstruction(models.Model):
    text = RichTextField(null=True, blank=True)
    index = models.PositiveIntegerField(default=1)
    blueprint = models.ForeignKey(
        DonationBlueprint,
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )
    zip_blueprint = models.ForeignKey(
        ZippedBlueprint,
        null=True,
        blank=True,
        on_delete=models.CASCADE
    )

    class Meta:
        ordering = ['index']
        constraints = [
            models.UniqueConstraint(
                fields=['index', 'blueprint'],
                name='unique_index_per_blueprint'
            ),
            models.UniqueConstraint(
                fields=['index', 'zip_blueprint'],
                name='unique_index_per_zipblueprint'
            ),
        ]

    def clean(self):
        if not self.blueprint and not self.zip_blueprint:
            raise ValidationError(
                'Must be linked to either a DonationBlueprint or '
                'a ZippedBlueprint.'
            )
        if self.blueprint and self.zip_blueprint:
            raise ValidationError(
                'Must be linked to either a DonationBlueprint or '
                'a ZippedBlueprint, but not both.'
            )
        super().clean()



# Generated by Django 3.2.13 on 2022-06-17 12:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ddm', '0005_alter_donationproject_public_key'),
    ]

    operations = [
        migrations.AlterField(
            model_name='datadonation',
            name='data',
            field=models.BinaryField(),
        ),
        migrations.AlterField(
            model_name='questionnaireresponse',
            name='data',
            field=models.BinaryField(),
        ),
    ]



from django.test import SimpleTestCase
from django.urls import reverse, resolve
from ddm.views import participation_flow


class TestParticipationFlowUrls(SimpleTestCase):

    def test_project_entry_url_resolves(self):
        url = reverse('project-entry', args=['project-slug'])
        self.assertEqual(resolve(url).func.view_class, participation_flow.EntryView)

    def test_project_datadonation_url_resolves(self):
        url = reverse('data-donation', args=['project-slug'])
        self.assertEqual(resolve(url).func.view_class, participation_flow.DataDonationView)

    def test_project_questionnaire_url_resolves(self):
        url = reverse('questionnaire', args=['project-slug'])
        self.assertEqual(resolve(url).func.view_class, participation_flow.QuestionnaireView)

    def test_project_exit_url_resolves(self):
        url = reverse('project-exit', args=['project-slug'])
        self.assertEqual(resolve(url).func.view_class, participation_flow.ExitView)
